package ru.dbotthepony.kbox2d.api

import ru.dbotthepony.kstarbound.math.Vector2d

/// The features that intersect to form the contact point
/// This must be 4 bytes or less. ?
data class ContactFeature(
	var indexA: Int = 0, ///< Feature index on shapeA
	var indexB: Int = 0, ///< Feature index on shapeB
	var typeA: Type = Type.VERTEX, ///< The feature type on shapeA
	var typeB: Type = Type.VERTEX, ///< The feature type on shapeB
) {
	enum class Type {
		VERTEX,
		FACE,
	}
}

/// Contact ids to facilitate warm starting.
data class ContactID(
	var key: Int = 0, ///< Used to quickly compare contact ids.
	val cf: ContactFeature = ContactFeature(),
)

/**
 * A manifold point is a contact point belonging to a contact
 * manifold. It holds details related to the geometry and dynamics
 * of the contact points.
 *
 * The local point usage depends on the manifold type:
 * - e_circles: the local center of circleB
 * - e_faceA: the local center of cirlceB or the clip point of polygonB
 * - e_faceB: the clip point of polygonA
 *
 * This structure is stored across time steps, so we keep it small.
 *
 * Note: the impulses are used for internal caching and may not
 * provide reliable contact forces, especially for high speed collisions.
 */
data class ManifoldPoint(
	var localPoint: Vector2d = Vector2d.ZERO, ///< usage depends on manifold type
	var normalImpulse: Double = 0.0, ///< the non-penetration impulse
	var tangentImpulse: Double = 0.0, ///< the friction impulse
	var id: ContactID = ContactID(), ///< uniquely identifies a contact point between two shapes
)

/**
 * A manifold for two touching convex shapes.
 *
 * Box2D supports multiple types of contact:
 * - clip point versus plane with radius
 * - point versus point with radius (circles)
 *
 * The local point usage depends on the manifold type:
 * - e_circles: the local center of circleA
 * - e_faceA: the center of faceA
 * - e_faceB: the center of faceB
 *
 * Similarly the local normal usage:
 * - e_circles: not used
 * - e_faceA: the normal on polygonA
 * - e_faceB: the normal on polygonB
 *
 * We store contacts in this way so that position correction can
 * account for movement, which is critical for continuous physics.
 *
 * All contact scenarios must be expressed in one of these types.
 *
 * This structure is stored across time steps, so we keep it small.
 */
data class Manifold(
	val localNormal: Vector2d = Vector2d.ZERO,
	val localPoint: Vector2d = Vector2d.ZERO,
	val type: Type? = null,
	val points: List<ManifoldPoint> = listOf(),
) {
	enum class Type {
		CIRCLES,
		FACE_A,
		FACE_B,
	}

	companion object {
		val EMPTY = Manifold()
	}
}

interface IWorldManifold {
	val normal: Vector2d
	val points: Array<Vector2d>
	val separations: DoubleArray
}

/// This is used for determining the state of contact points.
enum class PointState {
	NULL, ///< point does not exist
	ADD, ///< point was added in the update
	PERSIST, ///< point persisted across the update
	REMOVE ///< point was removed in the update
}

/// Used for computing contact manifolds.
data class ClipVertex(
	var v: Vector2d = Vector2d.ZERO,
	var id: ContactID = ContactID(),
)

/// Ray-cast input data. The ray extends from p1 to p1 + maxFraction * (p2 - p1).
data class RayCastInput(
	val p1: Vector2d,
	val p2: Vector2d,
	val maxFraction: Double,
)

/// Ray-cast output data. The ray hits at p1 + fraction * (p2 - p1), where p1 and p2
/// come from b2RayCastInput.
data class RayCastOutput(
	// В оригинале этого нет, но для un-C++шивания кода, оно должно быть тут
	val hit: Boolean,
	val normal: Vector2d = Vector2d.ZERO,
	val fraction: Double = 1.0,
) {
	companion object {
		val MISS = RayCastOutput(false)
	}
}
