package ru.dbotthepony.kbox2d.api

import kotlin.math.sqrt

/// Friction mixing law. The idea is to allow either fixture to drive the friction to zero.
/// For example, anything slides on ice.
fun b2MixFriction(friction1: Double, friction2: Double): Double {
	return sqrt(friction1 * friction2)
}

/// Restitution mixing law. The idea is allow for anything to bounce off an inelastic surface.
/// For example, a superball bounces on anything.
fun b2MixRestitution(restitution1: Double, restitution2: Double): Double {
	return if (restitution1 > restitution2) restitution1 else restitution2
}

/// Restitution mixing law. This picks the lowest value.
fun b2MixRestitutionThreshold(threshold1: Double, threshold2: Double): Double {
	return if (threshold1 < threshold2) threshold1 else threshold2
}

data class ContactRegister(
	val primary: Boolean,
)

/**
 * A contact edge is used to connect bodies and contacts together
 * in a contact graph where each body is a node and each contact
 * is an edge. A contact edge belongs to a doubly linked list
 * maintained in each attached body. Each contact has two contact
 * nodes, one for each attached body.
 */
data class ContactEdge(
	val other: ru.dbotthepony.kbox2d.api.IBody, ///< provides quick access to the other body attached.
	val contact: IContact, ///< the contact
	var prev: ContactEdge? = null, ///< the previous contact edge in the body's contact list
	var next: ContactEdge? = null, ///< the next contact edge in the body's contact list
)

enum class ContactFlags(val bitmask: Int) {
	ISLAND(0x1),
	TOUCHING(0x2),
	ENABLED(0x4),
	FILTER(0x8),
	BULLET_HIT(0x10),
	TOI(0x20);

	fun isit(other: Int): Boolean {
		return (other and bitmask) == bitmask
	}

	fun update(other: Int, state: Boolean): Int {
		if (state)
			return or(other)
		else
			return not(other)
	}

	fun and(other: Int): Int {
		return other and bitmask
	}

	fun or(other: Int): Int {
		return other or bitmask
	}

	fun not(other: Int): Int {
		return other and bitmask.inv()
	}
}

/**
 * The class manages contact between two shapes. A contact exists for each overlapping
 * AABB in the broad-phase (except if filtered). Therefore a contact object may exist
 * that has no contact points.
 */
interface IContact {
	/// Get the contact manifold. Do not modify the manifold unless you understand the
	/// internals of Box2D.
	val manifold: Manifold

	/// Get the world manifold.
	val worldManifold: IWorldManifold

	/// Is this contact touching?
	val isTouching: Boolean

	/// Enable/disable this contact. This can be used inside the pre-solve
	/// contact listener. The contact is only disabled for the current
	/// time step (or sub-step in continuous collisions).
	var isEnabled: Boolean

	/// Get the next contact in the world's contact list.
	val next: IContact?
	val prev: IContact?

	/// Get fixture A in this contact.
	val fixtureA: IFixture

	/// Get the child primitive index for fixture A.
	val childIndexA: Int

	/// Get fixture B in this contact.
	val fixtureB: IFixture

	/// Get the child primitive index for fixture B.
	val childIndexB: Int

	/// Override the default friction mixture. You can call this in b2ContactListener::PreSolve.
	/// This value persists until set or reset.
	/// Get the friction.
	var friction: Double

	/// Reset the friction mixture to the default value.
	fun resetFriction() {
		friction = b2MixFriction(fixtureA.friction, fixtureB.friction)
	}

	fun flagForFiltering()

	/// Override the default restitution mixture. You can call this in b2ContactListener::PreSolve.
	/// The value persists until you set or reset.
	var restitution: Double

	/// Reset the restitution to the default value.
	fun resetRestitution() {
		restitution = b2MixRestitution(fixtureA.restitution, fixtureB.restitution)
	}

	/// Override the default restitution velocity threshold mixture. You can call this in b2ContactListener::PreSolve.
	/// The value persists until you set or reset.
	/// Get the restitution threshold.
	var restitutionThreshold: Double

	/// Reset the restitution threshold to the default value.
	fun resetRestitutionThreshold() {
		restitutionThreshold = b2MixRestitutionThreshold(fixtureA.restitutionThreshold, fixtureB.restitutionThreshold)
	}

	/// Set the desired tangent speed for a conveyor belt behavior. In meters per second.
	/// Get the desired tangent speed. In meters per second.
	var tangentSpeed: Double

	/// Evaluate this contact with your own manifold and transforms.
	fun evaluate(xfA: Transform, xfB: Transform): Manifold
}
