package ru.dbotthepony.kbox2d.api

import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Vector2d

sealed interface IFilter {
	/**
	 * The collision category bits. Normally you would just set one bit.
	 */
	val categoryBits: Long

	/**
	 * The collision mask bits. This states the categories that this
	 * shape would accept for collision.
	 */
	val maskBits: Long

	/**
	 * Collision groups allow a certain group of objects to never collide (negative)
	 * or always collide (positive). Zero means no collision group. Non-zero group
	 * filtering always wins against the mask bits.
	 */
	val groupIndex: Int
}

data class Filter(
	override var categoryBits: Long = 0x1L,
	override var maskBits: Long = 0xFFFFL,
	override var groupIndex: Int = 0,
) : IFilter {
	fun immutable(): ImmutableFilter {
		return ImmutableFilter(categoryBits, maskBits, groupIndex)
	}
}

data class ImmutableFilter(
	override val categoryBits: Long,
	override val maskBits: Long,
	override val groupIndex: Int,
) : IFilter

data class FixtureDef(
	/// The shape, this must be set. The shape will be cloned.
	var shape: IShape<*>? = null,

	var friction: Double = 0.2,

	/// The restitution (elasticity) usually in the range [0,1].
	var restitution: Double = 0.0,

	/// Restitution velocity threshold, usually in m/s. Collisions above this
	/// speed have restitution applied (will bounce).
	var restitutionThreshold: Double = 1.0 * b2_lengthUnitsPerMeter,

	/// The density, usually in kg/m^2.
	var density: Double = 0.0,

	/// A sensor shape collects contact information but never generates a collision
	/// response.
	var isSensor: Boolean = false,

	/// Use this to store application specific fixture data.
	var userData: Any? = null,

	val filter: Filter = Filter()
)

data class FixtureProxy(
	var aabb: AABB,
	val fixture: IFixture,
	val childIndex: Int,
) {
	private var setProxyID = false

	var proxyId: Int = e_nullProxy
		set(value) {
			if (!setProxyID) {
				field = value
				setProxyID = true
				return
			}

			throw IllegalStateException("FixtureProxy should be immutable (tried to set value $value, already having $field)")
		}
}

/**
 * A fixture is used to attach a shape to a body for collision detection. A fixture
 * inherits its transform from its parent. Fixtures hold additional non-geometric data
 * such as friction, collision filters, etc.
 * Fixtures are created via b2Body::CreateFixture.
 * @warning you cannot reuse fixtures.
 */
interface IFixture {
	/// Get the type of the child shape. You can use this to down cast to the concrete shape.
	/// @return the shape type.
	val type: IShape.Type

	/// Get the child shape. You can modify the child shape, however you should not change the
	/// number of vertices because this will crash some collision caching mechanisms.
	/// Manipulating the shape may lead to non-physical behavior.
	val shape: IShape<*>

	/// Set if this fixture is a sensor.
	/// Is this fixture a sensor (non-solid)?
	/// @return the true if the shape is a sensor.
	var isSensor: Boolean

	/// Set the contact filtering data. This will not update contacts until the next time
	/// step when either parent body is active and awake.
	/// This automatically calls Refilter.
	var filter: IFilter

	/// Call this if you want to establish collision that was previously disabled by b2ContactFilter::ShouldCollide.
	fun refilter()

	/// Get the parent body of this fixture. This is nullptr if the fixture is not attached.
	/// @return the parent body.
	val body: ru.dbotthepony.kbox2d.api.IBody?

	/// Get the next fixture in the parent body's fixture list.
	/// @return the next shape.
	val next: IFixture?

	/// Get the user data that was assigned in the fixture definition. Use this to
	/// store your application specific data.
	val userData: Any?

	/// Test a point for containment in this fixture.
	/// @param p a point in world coordinates.
	fun testPoint(p: Vector2d): Boolean {
		return shape.testPoint(checkNotNull(body) { "Tried to use detached fixture" }.transform, p)
	}

	/// Cast a ray against this shape.
	/// @param output the ray-cast results.
	/// @param input the ray-cast input parameters.
	/// @param childIndex the child shape index (e.g. edge index)
	fun rayCast(input: RayCastInput, childIndex: Int): RayCastOutput {
		return shape.rayCast(input, checkNotNull(body) { "Tried to use detached fixture" }.transform, childIndex)
	}

	/// Get the mass data for this fixture. The mass data is based on the density and
	/// the shape. The rotational inertia is about the shape's origin. This operation
	/// may be expensive.
	fun getMassData(): MassData {
		return shape.computeMass(density)
	}

	/// Set the density of this fixture. This will _not_ automatically adjust the mass
	/// of the body. You must call b2Body::ResetMassData to update the body's mass.
	/// Get the density of this fixture.
	var density: Double

	/// Get the coefficient of friction.
	/// Set the coefficient of friction. This will _not_ change the friction of
	/// existing contacts.
	var friction: Double

	/// Get the coefficient of restitution.
	/// Set the coefficient of restitution. This will _not_ change the restitution of
	/// existing contacts.
	var restitution: Double

	/// Get the restitution velocity threshold.
	/// Set the restitution threshold. This will _not_ change the restitution threshold of
	/// existing contacts.
	var restitutionThreshold: Double

	/// Get the fixture's AABB. This AABB may be enlarge and/or stale.
	/// If you need a more accurate AABB, compute it using the shape and
	/// the body transform.
	fun getAABB(childIndex: Int): AABB {
		return proxies[childIndex].aabb
	}

	/// Dump this fixture to the log file. (Log4J)
	fun dump(childIndex: Int) { }

	val proxies: List<FixtureProxy>
}
