package ru.dbotthepony.kbox2d.api

import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kbox2d.dynamics.joint.AbstractJoint
import ru.dbotthepony.kstarbound.math.Vector2d
import kotlin.math.PI

data class StiffnessResult(val stiffness: Double, val damping: Double)

/**
 * Utility to compute linear stiffness values from frequency and damping ratio
 */
fun b2LinearStiffness(
	frequencyHertz: Double,
	dampingRatio: Double,
	bodyA: IBody?,
	bodyB: IBody?,
): StiffnessResult {
	val massA = bodyA?.mass ?: 0.0
	val massB = bodyB?.mass ?: 0.0
	val mass: Double

	if (massA > 0.0 && massB > 0.0) {
		mass = massA * massB / (massA + massB)
	} else if (massA > 0.0) {
		mass = massA
	} else {
		mass = massB
	}

	val omega = 2.0 * PI * frequencyHertz

	return StiffnessResult(
		stiffness = mass * omega * omega,
		damping = 2.0 * mass * dampingRatio * omega
	)
}

/**
 * Utility to compute rotational stiffness values frequency and damping ratio
 */
fun b2AngularStiffness(
	frequencyHertz: Double,
	dampingRatio: Double,
	bodyA: IBody?,
	bodyB: IBody?,
): StiffnessResult {
	val inertiaA = bodyA?.inertia ?: 0.0
	val inertiaB = bodyB?.inertia ?: 0.0
	val inertia: Double

	if (inertiaA > 0.0 && inertiaB > 0.0) {
		inertia = inertiaA * inertiaB / (inertiaA + inertiaB)
	} else if (inertiaA > 0.0) {
		inertia = inertiaA
	} else {
		inertia = inertiaB
	}

	val omega = 2.0 * PI * frequencyHertz

	return StiffnessResult(
		stiffness = inertia * omega * omega,
		damping = 2.0 * inertia * dampingRatio * omega
	)
}

enum class JointType {
	REVOLUTE,
	PRISMATIC,
	DISTANCE,
	PULLEY,
	MOUSE,
	GEAR,
	WHEEL,
	WELD,
	FRICTION,
	MOTOR
}

data class Jacobian(
	val linear: Vector2d,
	val angularA: Double,
	val angularB: Double
)

/**
 * A joint edge is used to connect bodies and joints together
 * in a joint graph where each body is a node and each joint
 * is an edge. A joint edge belongs to a doubly linked list
 * maintained in each attached body. Each joint has two joint
 * nodes, one for each attached body.
 */
class JointEdge(
	other: IBody?, ///< provides quick access to the other body attached.
	val joint: IJoint, ///< the joint
	var prev: JointEdge? = null, ///< the previous joint edge in the body's joint list
	var next: JointEdge? = null ///< the next joint edge in the body's joint list
) {
	val otherNullable: IBody? = other
	val other: IBody get() = checkNotNull(otherNullable) { "Other body is not present" }
}

sealed interface IJointDef {
	/**
	 * The joint type is set automatically for concrete joint types.
	 */
	val type: JointType

	/**
	 * The first attached body.
	 */
	val bodyA: IBody?

	/**
	 * The second attached body.
	 */
	val bodyB: IBody?

	/**
	 * Set this flag to true if the attached bodies should collide.
	 */
	var collideConnected: Boolean

	/**
	 * Use this to attach application specific data to your joints.
	 */
	var userData: Any?
}

class DistanceJointDef(
	b1: IBody,
	b2: IBody,
	anchor1: Vector2d,
	anchor2: Vector2d
) : IJointDef {
	override val type: JointType = JointType.DISTANCE

	/**
	 * The rest length of this joint. Clamped to a stable minimum value.
	 */
	var length: Double

	/**
	 * Minimum length. Clamped to a stable minimum value.
	 */
	var minLength: Double

	/**
	 * Maximum length. Must be greater than or equal to the minimum length.
	 */
	var maxLength: Double

	/**
	 * The linear stiffness in N/m.
	 */
	val stiffness: Double = 0.0

	/**
	 * The linear damping in N*s/m.
	 */
	val damping: Double = 0.0

	override var bodyA: IBody = b1
	override var bodyB: IBody = b2

	override var collideConnected: Boolean = false
	override var userData: Any? = null

	/**
	 * The local anchor point relative to bodyA's origin.
	 */
	var localAnchorA: Vector2d = bodyA.getLocalPoint(anchor1)

	/**
	 * The local anchor point relative to bodyB's origin.
	 */
	var localAnchorB: Vector2d = bodyB.getLocalPoint(anchor2)

	init {
		val d = anchor2 - anchor1
		length = b2Max(d.length, b2_linearSlop)
		minLength = length
		maxLength = length
	}
}

class RevoluteJointDef(
	b1: IBody,
	b2: IBody,
	anchor: Vector2d,
) : IJointDef {
	override val type: JointType = JointType.REVOLUTE
	override var bodyA: IBody = b1
	override var bodyB: IBody = b2
	override var collideConnected: Boolean = false
	override var userData: Any? = null

	/**
	 * The local anchor point relative to bodyA's origin.
	 */
	var localAnchorA = bodyA.getLocalPoint(anchor)

	/**
	 * The local anchor point relative to bodyB's origin.
	 */
	var localAnchorB = bodyB.getLocalPoint(anchor)

	/**
	 * The bodyB angle minus bodyA angle in the reference state (radians).
	 */
	var referenceAngle = bodyB.angle - bodyA.angle

	/**
	 * A flag to enable joint limits.
	 */
	var enableLimit = false

	/**
	 * The lower angle for the joint limit (radians).
	 */
	var lowerAngle = 0.0

	/**
	 * The upper angle for the joint limit (radians).
	 */
	var upperAngle = 0.0

	/**
	 * A flag to enable the joint motor.
	 */
	var enableMotor = false

	/**
	 * The desired motor speed. Usually in radians per second.
	 */
	var motorSpeed = 0.0

	/**
	 * The maximum motor torque used to achieve the desired motor speed.
	 *
	 * Usually in N-m.
	 */
	var maxMotorTorque = 0.0
}

class PrismaticJointDef(
	b1: IBody,
	b2: IBody,
	anchor: Vector2d,
	axis: Vector2d,
) : IJointDef {
	override val type: JointType = JointType.PRISMATIC
	override var bodyA: IBody = b1
	override var bodyB: IBody = b2
	override var collideConnected: Boolean = false
	override var userData: Any? = null

	/**
	 * The local anchor point relative to bodyA's origin.
	 */
	var localAnchorA = bodyA.getLocalPoint(anchor)

	/**
	 * The local anchor point relative to bodyB's origin.
	 */
	var localAnchorB = bodyB.getLocalPoint(anchor)

	/**
	 * The local translation unit axis in bodyA.
	 */
	var localAxisA = bodyA.getLocalVector(axis)

	/**
	 * The constrained angle between the bodies: bodyB_angle - bodyA_angle.
	 */
	var referenceAngle = bodyB.angle - bodyA.angle

	/**
	 * Enable/disable the joint limit.
	 */
	var enableLimit = false

	/**
	 * The lower translation limit, usually in meters.
	 */
	var lowerTranslation = 0.0

	/**
	 * The upper translation limit, usually in meters.
	 */
	var upperTranslation = 0.0

	/**
	 * Enable/disable the joint motor.
	 */
	var enableMotor = false

	/**
	 * The maximum motor torque, usually in N-m.
	 */
	var maxMotorForce = 0.0

	/**
	 * The desired motor speed in radians per second.
	 */
	var motorSpeed = 0.0
}

/**
 * Pulley joint definition. This requires two ground anchors,
 * two dynamic body anchor points, and a pulley ratio.
 */
class PulleyJointDef(
	b1: IBody,
	b2: IBody,

	/**
	 * The first ground anchor in world coordinates. This point never moves.
	 */
	var groundAnchorA: Vector2d,

	/**
	 * The second ground anchor in world coordinates. This point never moves.
	 */
	var groundAnchorB: Vector2d,

	anchorA: Vector2d,
	anchorB: Vector2d,
	ratio: Double,
) : IJointDef {
	override val type: JointType = JointType.PULLEY
	override var bodyA: IBody = b1
	override var bodyB: IBody = b2
	override var collideConnected: Boolean = false
	override var userData: Any? = null

	/**
	 * The pulley ratio, used to simulate a block-and-tackle.
	 */
	var ratio: Double = ratio
		set(value) {
			require(value > b2_epsilon) { "Ratio is too small: $value" }
			field = value
		}

	init {
		// KBox2D: trigger sanity check at least once
		this.ratio = ratio
	}

	/**
	 * The local anchor point relative to bodyA's origin.
	 */
	var localAnchorA = bodyA.getLocalPoint(anchorA)

	/**
	 * The local anchor point relative to bodyB's origin.
	 */
	var localAnchorB = bodyB.getLocalPoint(anchorB)

	/**
	 * The a reference length for the segment attached to bodyA.
	 */
	var lengthA: Double = (anchorA - groundAnchorA).length

	/**
	 * The b reference length for the segment attached to bodyB.
	 */
	var lengthB: Double = (anchorB - groundAnchorB).length
}

/**
 * Gear joint definition. This definition requires two existing
 * revolute or prismatic joints (any combination will work).
 * @warning bodyB on the input joints must both be dynamic
 */
class GearJointDef(
	override var bodyA: IBody,
	override var bodyB: IBody,

	/**
	 * The first revolute/prismatic joint attached to the gear joint.
	 */
	var joint1: AbstractJoint,

	/**
	 * The second revolute/prismatic joint attached to the gear joint.
	 */
	var joint2: AbstractJoint,

	/**
	 * The gear ratio.
	 * @see ru.dbotthepony.kbox2d.dynamics.joint.GearJoint for explanation.
	 */
	var ratio: Double,
) : IJointDef {
	override val type: JointType = JointType.GEAR
	override var collideConnected: Boolean = false
	override var userData: Any? = null
}

/**
 * Mouse joint definition. This requires a world target point,
 * tuning parameters, and the time step.
 *
 * KBox2D: In Box2D, Body A is not used and you put meaningless data into it just to not make it crash.
 *
 * In KBox2D, Body A is also meaningless, but you are not required to put anything it (leave as null).
 *
 * Putting anything in Body A or leaving it as null has no effect, other than bloating body A joint linked list.
 */
class MouseJointDef(
	/**
	 * The initial world target point. This is assumed
	 * to coincide with the body anchor initially.
     */
	var target: Vector2d,

	override var bodyB: IBody,
	override val bodyA: IBody? = null,

	/**
	 * The maximum constraint force that can be exerted
	 * to move the candidate body. Usually you will express
	 * as some multiple of the weight (multiplier * mass * gravity).
	 */
	var maxForce: Double = 0.0,
	var stiffness: Double = 0.0,
	var damping: Double = 0.0,
) : IJointDef {
	override val type: JointType = JointType.MOUSE

	override var collideConnected: Boolean = false
	override var userData: Any? = null

	fun linearStiffness(
		frequencyHertz: Double,
		dampingRatio: Double,
		bodyA: IBody?,
		bodyB: IBody?
	): MouseJointDef {
		val (stiffness, damping) = b2LinearStiffness(frequencyHertz, dampingRatio, bodyA, bodyB)
		this.stiffness = stiffness
		this.damping = damping
		return this
	}

	fun angularStiffness(
		frequencyHertz: Double,
		dampingRatio: Double,
		bodyA: IBody?,
		bodyB: IBody?
	): MouseJointDef {
		val (stiffness, damping) = b2AngularStiffness(frequencyHertz, dampingRatio, bodyA, bodyB)
		this.stiffness = stiffness
		this.damping = damping
		return this
	}
}

/**
 * Wheel joint definition. This requires defining a line of
 * motion using an axis and an anchor point. The definition uses local
 * anchor points and a local axis so that the initial configuration
 * can violate the constraint slightly. The joint translation is zero
 * when the local anchor points coincide in world space. Using local
 * anchors and a local axis helps when saving and loading a game.
 */
class WheelJointDef(
	override var bodyA: IBody,
	override var bodyB: IBody,
	anchor: Vector2d,
	axis: Vector2d,

	/**
	 * Enable/disable the joint limit.
	 */
	var enableLimit: Boolean = false,

	/**
	 * The lower translation limit, usually in meters.
	 */
	var lowerTranslation: Double = 0.0,

	/**
	 * The upper translation limit, usually in meters.
	 */
	var upperTranslation: Double = 0.0,

	/**
	 * Enable/disable the joint motor.
	 */
	var enableMotor: Boolean = false,

	/**
	 * The maximum motor torque, usually in N-m.
	 */
	var maxMotorTorque: Double = 0.0,

	/**
	 * The desired motor speed in radians per second.
	 */
	var motorSpeed: Double = 0.0,

	/**
	 * Suspension stiffness. Typically in units N/m.
	 */
	var stiffness: Double = 0.0,

	/**
	 * Suspension damping. Typically in units of N*s/m.
	 */
	var damping: Double = 0.0,
) : IJointDef {
	override val type: JointType = JointType.WHEEL
	override var collideConnected: Boolean = false
	override var userData: Any? = null

	/**
	 * The local anchor point relative to bodyA's origin.
	 */
	var localAnchorA: Vector2d = bodyA.getLocalPoint(anchor)

	/**
	 * The local anchor point relative to bodyB's origin.
	 */
	var localAnchorB: Vector2d = bodyB.getLocalPoint(anchor)

	/**
	 * The local translation axis in bodyA.
	 */
	var localAxisA: Vector2d = bodyA.getLocalVector(axis)

	fun linearStiffness(
		frequencyHertz: Double,
		dampingRatio: Double,
		bodyA: IBody = this.bodyA,
		bodyB: IBody = this.bodyB
	): WheelJointDef {
		val (stiffness, damping) = b2LinearStiffness(frequencyHertz, dampingRatio, bodyA, bodyB)
		this.stiffness = stiffness
		this.damping = damping
		return this
	}

	fun angularStiffness(
		frequencyHertz: Double,
		dampingRatio: Double,
		bodyA: IBody = this.bodyA,
		bodyB: IBody = this.bodyB
	): WheelJointDef {
		val (stiffness, damping) = b2AngularStiffness(frequencyHertz, dampingRatio, bodyA, bodyB)
		this.stiffness = stiffness
		this.damping = damping
		return this
	}
}

/**
 * Weld joint definition. You need to specify local anchor points
 * where they are attached and the relative body angle. The position
 * of the anchor points is important for computing the reaction torque.
 */
class WeldJointDef(
	override var bodyA: IBody,
	override var bodyB: IBody,
	anchor: Vector2d,

	/**
	 * Suspension stiffness. Typically in units N/m.
	 */
	var stiffness: Double = 0.0,

	/**
	 * Suspension damping. Typically in units of N*s/m.
	 */
	var damping: Double = 0.0,
) : IJointDef {
	override val type: JointType = JointType.WELD
	override var collideConnected: Boolean = false
	override var userData: Any? = null

	/**
	 * The bodyB angle minus bodyA angle in the reference state (radians).
	 */
	var referenceAngle: Double = bodyB.angle - bodyA.angle

	/**
	 * The local anchor point relative to bodyA's origin.
	 */
	var localAnchorA: Vector2d = bodyA.getLocalPoint(anchor)

	/**
	 * The local anchor point relative to bodyB's origin.
	 */
	var localAnchorB: Vector2d = bodyB.getLocalPoint(anchor)

	fun linearStiffness(
		frequencyHertz: Double,
		dampingRatio: Double,
		bodyA: IBody = this.bodyA,
		bodyB: IBody = this.bodyB
	): WeldJointDef {
		val (stiffness, damping) = b2LinearStiffness(frequencyHertz, dampingRatio, bodyA, bodyB)
		this.stiffness = stiffness
		this.damping = damping
		return this
	}

	fun angularStiffness(
		frequencyHertz: Double,
		dampingRatio: Double,
		bodyA: IBody = this.bodyA,
		bodyB: IBody = this.bodyB
	): WeldJointDef {
		val (stiffness, damping) = b2AngularStiffness(frequencyHertz, dampingRatio, bodyA, bodyB)
		this.stiffness = stiffness
		this.damping = damping
		return this
	}
}

/**
 * Friction joint definition.
 */
class FrictionJointDef(
	override var bodyA: IBody,
	override var bodyB: IBody,
	anchor: Vector2d,

	/**
	 * The maximum friction force in N.
	 */
	var maxForce: Double = 0.0,

	/**
	 * The maximum friction torque in N-m.
	 */
	var maxTorque: Double = 0.0,

	override var collideConnected: Boolean = false,
	override var userData: Any? = null
) : IJointDef {
	override val type: JointType = JointType.FRICTION

	/**
	 * The local anchor point relative to bodyA's origin.
	 */
	var localAnchorA: Vector2d = bodyA.getLocalPoint(anchor)

	/**
	 * The local anchor point relative to bodyB's origin.
	 */
	var localAnchorB: Vector2d = bodyB.getLocalPoint(anchor)
}

class MotorJointDef(
	override var bodyA: IBody,
	override var bodyB: IBody,

	/**
	 * The maximum motor force in N.
	 */
	var maxForce: Double = 1.0,

	/**
	 * The maximum motor torque in N-m.
	 */
	var maxTorque: Double = 1.0,

	override var collideConnected: Boolean = false,
	override var userData: Any? = null
) : IJointDef {
	override val type: JointType = JointType.MOTOR

	/**
	 * Position of bodyB minus the position of bodyA, in bodyA's frame, in meters.
	 */
	var linearOffset: Vector2d = bodyA.getLocalPoint(bodyB.position)

	/**
	 * The bodyB angle minus bodyA angle in radians.
	 */
	var angularOffset: Double = bodyB.angle - bodyA.angle

	/**
	 * Position correction factor in the range [0,1].
	 */
	var correctionFactor: Double = 0.3
		set(value) {
			require(value in 0.0 .. 1.0) { "Invalid correction factor of $value" }
			field = value
		}
}

interface IJoint : IMovable {
	/**
	 * Get the type of the concrete joint.
	 */
	val type: JointType

	/**
	 * Get the first body attached to this joint.
	 */
	val bodyA: IBody

	/**
	 * Get the second body attached to this joint.
	 */
	val bodyB: IBody

	/**
	 * Get the anchor point on bodyA in world coordinates.
	 */
	val anchorA: Vector2d

	/**
	 * Get the anchor point on bodyB in world coordinates.
	 */
	val anchorB: Vector2d

	/**
	 * Get the reaction force on bodyB at the joint anchor in Newtons.
	 */
	fun getReactionForce(inv_dt: Double): Vector2d

	/**
	 * Get the reaction torque on bodyB in N*m.
	 */
	fun getReactionTorque(inv_dt: Double): Double

	/**
	 * Get the next joint the world joint list.
	 *
	 * Kotlin: This is not a list, but, as in C++ impl, a custom
	 * linked list.
	 */
	val next: IJoint?
	val prev: IJoint?

	fun destroy() {
		bodyA.world.destroyJoint(this)
	}

	/**
	 * Short-cut function to determine if either body is enabled.
	 */
	val isEnabled: Boolean get() = bodyA.isEnabled || bodyB.isEnabled

	/**
	 * Get collide connected.
	 * Note: modifying the collide connect flag won't work correctly because
	 * the flag is only checked when fixture AABBs begin to overlap.
	 */
	val collideConnected: Boolean

	/**
	 * Dump this joint to the log file.
	 */
	fun dump() { LOGGER.warn("Dump is not supported for this join type: $this") }

	/**
	 * Debug draw this joint
	 */
	fun draw(draw: IDebugDraw) { }

	var userData: Any?

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
