package ru.dbotthepony.kbox2d.collision.shapes

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Vector2d

/**
 * A chain shape is a free form sequence of line segments.
 * The chain has one-sided collision, with the surface normal pointing to the right of the edge.
 * This provides a counter-clockwise winding like the polygon shape.
 * Connectivity information is used to create smooth collisions.
 * @warning the chain will not collide properly if there are self-intersections.
 */
class ChainShape : IShape<ChainShape> {
	override fun copy(): ChainShape {
		return ChainShape().also {
			it.vertices.addAll(vertices)
			it.prevVertex = prevVertex
			it.nextVertex = nextVertex
			it.edgeCache = arrayOfNulls(vertices.size - 1)
			it.rayEdgeCache = arrayOfNulls(vertices.size)
		}
	}

	private fun validateInput(vertices: List<Vector2d>) {
		for (i in 1 until vertices.size) {
			val v1 = vertices[i - 1]
			val v2 = vertices[i]

			v1.isFiniteOrThrow { "Vertex at ${i - 1} is invalid" }
			v2.isFiniteOrThrow { "Vertex at $i is invalid" }

			require(b2DistanceSquared(v1, v2) > b2_linearSlop * b2_linearSlop) {
				"Vertices are too close together, at indices ${i - 1} ($v1) and $i ($v2)"
			}
		}
	}

	/**
	 * Create a loop. This automatically adjusts connectivity.
	 * @param vertices an array of vertices, these are copied
	 * @param count the vertex count
	 */
	fun createLoop(vertices: List<Vector2d>) {
		// b2Assert(m_vertices == nullptr && m_count == 0);
		// KBox2D: not required because we are much more flexible at memory than C++

		require(vertices.size >= 3) { "Can not create loop with ${vertices.size} vertices" }

		validateInput(vertices)
		this.vertices.clear()

		this.vertices.addAll(vertices)
		this.vertices.add(vertices[0])

		prevVertex = vertices[vertices.size - 1]
		nextVertex = vertices[1]

		edgeCache = arrayOfNulls(this.vertices.size - 1)
		rayEdgeCache = arrayOfNulls(this.vertices.size)
	}

	/**
	 * Create a chain with ghost vertices to connect multiple chains together.
	 * @param vertices an array of vertices, these are copied
	 * @param count the vertex count
	 * @param prevVertex previous vertex from chain that connects to the start
	 * @param nextVertex next vertex from chain that connects to the end
	 */
	fun createChain(vertices: List<Vector2d>, prevVertex: Vector2d, nextVertex: Vector2d) {
		// b2Assert(m_vertices == nullptr && m_count == 0);
		// KBox2D: not required because we are much more flexible at memory than C++

		require(vertices.size >= 2) { "Can not create chain with ${vertices.size} vertices" }

		validateInput(vertices)
		this.vertices.clear()
		this.vertices.addAll(vertices)

		this.prevVertex = prevVertex
		this.nextVertex = nextVertex

		edgeCache = arrayOfNulls(this.vertices.size - 1)
		rayEdgeCache = arrayOfNulls(this.vertices.size)
	}

	internal val vertices = ArrayList<Vector2d>()
	internal var prevVertex: Vector2d? = null
	internal var nextVertex: Vector2d? = null

	override val type: IShape.Type = IShape.Type.CHAIN
	// edge count = vertex count - 1
	override val childCount: Int get() = vertices.size - 1

	private var edgeCache: Array<EdgeShape?> = arrayOfNulls(0)
	private var rayEdgeCache: Array<EdgeShape?> = arrayOfNulls(0)

	/**
	 * Get a child edge.
	 */
	fun getChildEdge(index: Int): EdgeShape {
		val getEdge = edgeCache[index]

		if (getEdge != null) {
			return getEdge
		}

		val edge = EdgeShape()

		edge.vertex1 = vertices[index]
		edge.vertex2 = vertices[index + 1]
		edge.oneSided = true

		if (index > 0) {
			edge.vertex0 = vertices[index - 1]
		} else {
			edge.vertex0 = prevVertex ?: throw NullPointerException("prevVertex is null")
		}

		if (index < vertices.size - 2) {
			edge.vertex3 = vertices[index + 2]
		} else {
			edge.vertex3 = nextVertex ?: throw NullPointerException("nextVertex is null")
		}

		edgeCache[index] = edge
		return edge
	}

	override fun rayCast(input: RayCastInput, transform: Transform, childIndex: Int): RayCastOutput {
		val getEdge = rayEdgeCache[childIndex]

		if (getEdge != null) {
			return getEdge.rayCast(input, transform, 0)
		}

		val edgeShape = EdgeShape()

		val i1 = childIndex
		var i2 = childIndex + 1

		if (i2 == vertices.size) {
			i2 = 0
		}

		edgeShape.vertex1 = vertices[i1]
		edgeShape.vertex2 = vertices[i2]

		rayEdgeCache[childIndex] = edgeShape
		return edgeShape.rayCast(input, transform, 0)
	}

	override fun computeAABB(transform: Transform, childIndex: Int): AABB {
		val i1 = childIndex
		var i2 = childIndex + 1

		if (i2 == vertices.size) {
			i2 = 0
		}

		val v1 = b2Mul(transform, vertices[i1])
		val v2 = b2Mul(transform, vertices[i2])

		val lower = b2Min(v1, v2)
		val upper = b2Max(v1, v2)

		val r = Vector2d(radius, radius)

		return AABB(
			mins = lower - r,
			maxs = upper + r
		)
	}

	override fun computeMass(density: Double): MassData {
		return MASS
	}

	override var radius: Double = b2_polygonRadius
		set(value) {
			if (value == b2_polygonRadius) {
				field = value
			}

			throw IllegalArgumentException("")
		}

	companion object {
		private val MASS = MassData(mass = 0.0, center = Vector2d.ZERO, inertia = 0.0)
	}
}