package ru.dbotthepony.kbox2d.collision.shapes

import it.unimi.dsi.fastutil.ints.IntArrayList
import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Vector2d
import ru.dbotthepony.kstarbound.math.times

private const val inv3 = 1.0 / 3.0

private fun computeCentroid(vs: List<Vector2d>): Vector2d {
	require(vs.size >= 3) { "Got only ${vs.size} vertices" }

	var c = Vector2d.ZERO
	var area = 0.0

	// Get a reference point for forming triangles.
	// Use the first vertex to reduce round-off errors.
	val s = vs[0]

	for (i in vs.indices) {
		// Triangle vertices.
		val p1 = vs[0] - s;
		val p2 = vs[i] - s;
		val p3 = if (i + 1 < vs.size) vs[i + 1] - s else vs[0] - s;

		val e1 = p2 - p1;
		val e2 = p3 - p1;

		val D = b2Cross(e1, e2);

		val triangleArea = 0.5 * D;
		area += triangleArea;

		// Area weighted centroid
		c += triangleArea * inv3 * (p1 + p2 + p3);
	}

	// Centroid
	check(area > b2_epsilon) { area }

	return (1.0 / area) * c + s
}

class PolygonShape : IShape<PolygonShape> {
	override fun copy(): PolygonShape {
		return PolygonShape().also {
			it.centroid = centroid
			it.vertices.addAll(vertices)
			it.normals.addAll(normals)
		}
	}

	/**
	 * Create a convex hull from the given array of local points.
	 * The count must be in the range [3, b2_maxPolygonVertices].
	 * @warning the points may be re-ordered, even if they form a convex polygon
	 * @warning collinear points are handled but not removed. Collinear points
	 * may lead to poor stacking behavior.
	 */
	fun set(vertices: List<Vector2d>) {
		this.vertices.clear()
		this.normals.clear()

		require(vertices.size >= 3) { "Got only ${vertices.size} points" }

		// Perform welding and copy vertices into local buffer.
		val ps = ArrayList<Vector2d>()

		for ((i, v) in vertices.withIndex()) {
			var unique = true

			for ((j, other) in ps.withIndex()) {
				if (v.distanceSquared(other) < ((0.5 * b2_linearSlop) * (0.5 * b2_linearSlop))) {
					unique = false
					break
				}
			}

			if (unique) {
				ps.add(v)
			}
		}

		check(ps.size >= 3) { "Polygon is degenerate" }

		// Create the convex hull using the Gift wrapping algorithm
		// http://en.wikipedia.org/wiki/Gift_wrapping_algorithm

		// Find the right most point on the hull
		var i0 = 0
		var x0 = ps[0].x

		for (i in 1 until ps.size) {
			val x = ps[i].x

			if (x > x0 || (x == x0 && ps[i].y < ps[i0].y)) {
				i0 = i
				x0 = x
			}
		}

		val hull = IntArrayList()
		var ih = i0

		while (true) {
			hull.add(ih)

			var ie = 0
			for (j in 1 until ps.size) {
				if (ie == ih) {
					ie = j
					continue
				}

				val r = ps[ie] - ps[ih]
				val v = ps[j] - ps[ih]

				val c = b2Cross(r, v)

				if (c < 0.0) {
					ie = j
				}

				// Collinearity check
				if (c == 0.0 && v.lengthSquared > r.lengthSquared) {
					ie = j
				}
			}

			ih = ie

			if (ie == i0) {
				break
			}
		}

		check(hull.size >= 3) { "Polygon is degenerate" }

		// Copy vertices.
		for (i in hull.indices) {
			this.vertices.add(ps[hull.getInt(i)])
		}

		// Compute normals. Ensure the edges have non-zero length.
		for (i in hull.indices) {
			val i2 = if (i + 1 < hull.size) i + 1 else 0
			val edge = this.vertices[i2] - this.vertices[i]
			check(edge.lengthSquared > b2_epsilon * b2_epsilon)
			this.normals.add(b2Cross(edge, 1.0).normalized)
		}

		// Compute the polygon centroid.
		centroid = computeCentroid(this.vertices)
	}

	/**
	 * Build vertices to represent an axis-aligned box centered on the local origin.
	 * @param hx the half-width.
	 * @param hy the half-height.
	 */
	fun setAsBox(hx: Double, hy: Double) {
		vertices.clear()
		normals.clear()

		vertices.add(Vector2d(-hx, -hy))
		vertices.add(Vector2d(hx, -hy))
		vertices.add(Vector2d(hx, hy))
		vertices.add(Vector2d(-hx, hy))

		normals.add(Vector2d.DOWN)
		normals.add(Vector2d.RIGHT)
		normals.add(Vector2d.UP)
		normals.add(Vector2d.LEFT)

		centroid = Vector2d.ZERO
	}

	/**
	 * Build vertices to represent an oriented box.
	 * @param hx the half-width.
	 * @param hy the half-height.
	 * @param center the center of the box in local coordinates.
	 * @param angle the rotation of the box in local coordinates.
	 */
	fun setAsBox(hx: Double, hy: Double, center: Vector2d, angle: Double) {
		vertices.clear()
		normals.clear()

		vertices.add(Vector2d(-hx, -hy))
		vertices.add(Vector2d(hx, -hy))
		vertices.add(Vector2d(hx, hy))
		vertices.add(Vector2d(-hx, hy))

		normals.add(Vector2d.DOWN)
		normals.add(Vector2d.RIGHT)
		normals.add(Vector2d.UP)
		normals.add(Vector2d.LEFT)

		centroid = center

		val xf = Transform(center, angle)

		// Transform vertices and normals.
		for (i in 0 until 4) {
			vertices[i] = b2Mul(xf, vertices[i])
			normals[i] = b2Mul(xf.q, normals[i])
		}
	}

	override fun testPoint(transform: Transform, p: Vector2d): Boolean {
		val pLocal = b2MulT(transform.q, p - transform.p)

		for (i in 0 until vertices.size) {
			val dot = b2Dot(normals[i], pLocal - vertices[i])

			if (dot > 0.0) {
				return false
			}
		}

		return true
	}

	override val type: IShape.Type = IShape.Type.POLYGON
	override val childCount: Int = 1

	override fun rayCast(input: RayCastInput, transform: Transform, childIndex: Int): RayCastOutput {
		// Put the ray into the polygon's frame of reference.
		val p1 = b2MulT(transform.q, input.p1 - transform.p)
		val p2 = b2MulT(transform.q, input.p2 - transform.p)
		val d = p2 - p1

		var lower = 0.0
		var upper = input.maxFraction

		var index = -1

		for (i in 0 until vertices.size) {
			// p = p1 + a * d
			// dot(normal, p - v) = 0
			// dot(normal, p1 - v) + a * dot(normal, d) = 0

			val numerator = b2Dot(normals[i], vertices[i] - p1)
			val denominator = b2Dot(normals[i], d)

			if (denominator == 0.0) {
				if (numerator < 0.0) {
					return RayCastOutput.MISS
				}
			} else {
				// Note: we want this predicate without division:
				// lower < numerator / denominator, where denominator < 0
				// Since denominator < 0, we have to flip the inequality:
				// lower < numerator / denominator <==> denominator * lower > numerator.

				if (denominator < 0.0 && numerator < lower * denominator) {
					// Increase lower.
					// The segment enters this half-space.
					lower = numerator / denominator
					index = i
				} else if (denominator > 0.0 && numerator < upper * denominator) {
					// Decrease upper.
					// The segment exits this half-space.
					upper = numerator / denominator
				}
			}

			// The use of epsilon here causes the assert on lower to trip
			// in some cases. Apparently the use of epsilon was to make edge
			// shapes work, but now those are handled separately.
			//if (upper < lower - b2_epsilon)

			if (upper < lower) {
				return RayCastOutput.MISS
			}
		}

		check(lower in 0.0 .. input.maxFraction) { "$lower <=> ${input.maxFraction}!" }

		if (index >= 0) {
			return RayCastOutput(
				hit = true,
				fraction = lower,
				normal = b2Mul(transform.q, normals[index])
			)
		}

		return RayCastOutput.MISS
	}

	override fun computeAABB(transform: Transform, childIndex: Int): AABB {
		var lower = b2Mul(transform, vertices[0])
		var upper = lower

		for (i in 1 until vertices.size) {
			val v = b2Mul(transform, vertices[i])
			lower = b2Min(lower, v)
			upper = b2Max(upper, v)
		}

		val r = Vector2d(radius, radius)

		return AABB(
			mins = lower - r,
			maxs = upper + r
		)
	}

	override fun computeMass(density: Double): MassData {
		// Polygon mass, centroid, and inertia.
		// Let rho be the polygon density in mass per unit area.
		// Then:
		// mass = rho * int(dA)
		// centroid.x = (1/mass) * rho * int(x * dA)
		// centroid.y = (1/mass) * rho * int(y * dA)
		// I = rho * int((x*x + y*y) * dA)
		//
		// We can compute these integrals by summing all the integrals
		// for each triangle of the polygon. To evaluate the integral
		// for a single triangle, we make a change of variables to
		// the (u,v) coordinates of the triangle:
		// x = x0 + e1x * u + e2x * v
		// y = y0 + e1y * u + e2y * v
		// where 0 <= u && 0 <= v && u + v <= 1.
		//
		// We integrate u from [0,1-v] and then v from [0,1].
		// We also need to use the Jacobian of the transformation:
		// D = cross(e1, e2)
		//
		// Simplification: triangle centroid = (1/3) * (p1 + p2 + p3)
		//
		// The rest of the derivation is handled by computer algebra.

		check(vertices.size >= 3) { vertices.size }

		var center = Vector2d.ZERO
		var area = 0.0
		var I = 0.0

		// Get a reference point for forming triangles.
		// Use the first vertex to reduce round-off errors.
		var s = vertices[0]

		for (i in 0 until vertices.size) {
			// Triangle vertices.
			val e1 = vertices[i] - s;
			val e2 = if (i + 1 < vertices.size) vertices[i+1] - s else vertices[0] - s

			val D = b2Cross(e1, e2)

			val triangleArea = 0.5f * D;
			area += triangleArea;

			// Area weighted centroid
			center += triangleArea * inv3 * (e1 + e2);

			val ex1 = e1.x
			val ey1 = e1.y
			val ex2 = e2.x
			val ey2 = e2.y

			val intx2 = ex1*ex1 + ex2*ex1 + ex2*ex2;
			val inty2 = ey1*ey1 + ey2*ey1 + ey2*ey2;

			I += (0.25f * inv3 * D) * (intx2 + inty2);
		}

		check(area > b2_epsilon) { "Area is too small: $area" }

		center *= 1.0 / area
		val center2 = center + s
		val mass = density * area

		return MassData(
			// Total mass
			mass = mass,
			// Center of mass
			center = center2,
			// Inertia tensor relative to the local origin (point s).
			// Shift to center of mass then to original body origin.
			inertia = density * I + mass * (b2Dot(center2, center2) - b2Dot(center, center)),
		)
	}

	/**
	 * Validate convexity. This is a very time consuming operation.
	 * @returns true if valid
	 */
	fun validate(doThrow: Boolean = false): Boolean {
		for (i in 0 until vertices.size) {
			val i1 = i
			val i2 = if (i < vertices.size - 1) i1 + 1 else 0
			val p = vertices[i1]
			val e = vertices[i2] - p

			if (!p.isFinite) {
				if (doThrow) {
					throw IllegalStateException("Vertex at $i1 is not finite")
				}

				return false
			}

			if (!e.isFinite) {
				if (doThrow) {
					throw IllegalStateException("Vertex at $i2 is not finite")
				}

				return false
			}

			for (j in 0 until vertices.size) {
				if (j == i1 || j == i2) {
					continue
				}

				val v = vertices[j] - p
				val c = b2Cross(e, v)

				if (c < 0.0) {
					if (doThrow) {
						throw IllegalStateException("Vertex at $j form concave shape")
					}

					return false
				}
			}
		}

		return true
	}

	internal var centroid = Vector2d.ZERO
	internal val vertices = ArrayList<Vector2d>()
	internal val normals = ArrayList<Vector2d>()

	val count: Int get() = vertices.size

	override var radius: Double = b2_polygonRadius
		set(value) {
			if (value == b2_polygonRadius) {
				field = value
				return
			}

			throw UnsupportedOperationException("For polygonal shapes this must be b2_polygonRadius")
		}
}