package ru.dbotthepony.kbox2d.dynamics

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kstarbound.math.Vector2d

open class Body(def: BodyDef, world: IB2World) : IBody {
	private var _world: IB2World? = world

	final override val world: IB2World
		get() = _world ?: throw IllegalStateException("Tried to use removed body")

	internal var flags: Int = 0
	internal var isOnIsland = false
	internal var islandIndex: Int = 0

	internal fun unlink() {
		_world = null
	}

	init {
		def.validate()

		if (def.bullet) {
			flags = BodyFlags.BULLET.or(flags)
		}

		if (def.fixedRotation) {
			flags = BodyFlags.FIXED_ROTATION.or(flags)
		}

		if (def.allowSleep) {
			flags = BodyFlags.AUTO_SLEEP.or(flags)
		}

		if (def.awake && def.type != BodyType.STATIC) {
			flags = BodyFlags.AWAKE.or(flags)
		}

		if (def.enabled) {
			flags = BodyFlags.ENABLED.or(flags)
		}
	}

	internal val sweep: Sweep = Sweep()

	final override val transform: Transform = Transform(def.position, def.angle)
	val xf by this::transform

	override val position: Vector2d
		get() = transform.position
	override val angle: Double
		get() = sweep.a

	override val userData: Any? = def.userData

	internal var sleepTime: Double = 0.0
	internal var torque: Double = 0.0
	internal var force: Vector2d = Vector2d.ZERO

	override var linearVelocity: Vector2d = def.linearVelocity
		set(value) {
			if (type == BodyType.STATIC)
				return

			if (value.dotProduct(value) > 0.0)
				isAwake = true

			field = value
		}

	override var angularVelocity: Double = def.angularVelocity
		set(value) {
			if (type == BodyType.STATIC)
				return

			if (value * value > 0.0)
				isAwake = true

			field = value
		}

	override var fixtureList: Fixture? = null
		protected set

	protected var fixtureCount: Int = 0

	override var jointList: JointEdge? = null
		internal set
	override var contactEdge: ContactEdge? = null
		internal set

	override var next: IBody? = null
		internal set
	override var prev: IBody? = null
		internal set

	override var linearDamping: Double = def.linearDamping
	override var angularDamping: Double = def.angularDamping
	override var gravityScale: Double = def.gravityScale

	override var mass: Double = 0.0
		protected set

	internal var invMass: Double = 0.0

	/**
	 * Rotational inertia about the center of mass.
 	 */
	internal var rotInertia: Double = 0.0

	/**
	 * Rotational inertia about the center of mass.
	 */
	internal var rotInertiaInv: Double = 0.0

	internal var I by this::rotInertia
	internal var invI by this::rotInertiaInv

	override var isBullet: Boolean
		get() = BodyFlags.BULLET.isit(flags)
		set(value) { flags = BodyFlags.BULLET.update(flags, value) }

	override var isAwake: Boolean
		get() = BodyFlags.AWAKE.isit(flags)
		set(value) {
			if (type == BodyType.STATIC || BodyFlags.AWAKE.isit(flags) == value)
				return

			if (value) {
				flags = flags or BodyFlags.AWAKE.bitmask
				sleepTime = 0.0
			} else {
				flags = flags and BodyFlags.AWAKE.bitmask.inv()
				sleepTime = 0.0
				linearVelocity = Vector2d.ZERO
				angularVelocity = 0.0
				force = Vector2d.ZERO
				torque = 0.0
			}
		}

	override var isEnabled: Boolean
		get() = BodyFlags.ENABLED.isit(flags)
		set(value) {
			if (world.isLocked)
				throw ConcurrentModificationException()

			if (value == isEnabled)
				return

			if (value) {
				flags = BodyFlags.ENABLED.not(flags)

				// Create all proxies.
				val broadPhase = world.contactManager.broadPhase

				for (fixture in fixtureIterator) {
					(fixture as Fixture).createProxies(broadPhase, transform)
				}

				world.notifyNewContacts()
			} else {
				flags = BodyFlags.ENABLED.or(flags)

				// Destroy all proxies.
				val broadPhase = world.contactManager.broadPhase

				for (fixture in fixtureIterator) {
					(fixture as Fixture).destroyProxies(broadPhase)
				}

				// Destroy the attached contacts.
				for (edge in contactEdgeIterator) {
					world.contactManager.destroy(edge.contact)
				}

				contactEdge = null
			}
		}

	override var isFixedRotation: Boolean
		get() = BodyFlags.FIXED_ROTATION.isit(flags)
		set(value) {
			if (value == isFixedRotation)
				return

			flags = BodyFlags.FIXED_ROTATION.update(flags, value)

			angularVelocity = 0.0
			resetMassData()
		}

	override var allowAutoSleep: Boolean
		get() = BodyFlags.AUTO_SLEEP.isit(flags)
		set(value) {
			flags = BodyFlags.AUTO_SLEEP.update(flags, value)

			if (!value) {
				isAwake = true
			}
		}

	override var type: BodyType = def.type
		set(value) {
			if (world.isLocked)
				throw ConcurrentModificationException()

			field = value

			resetMassData()

			if (value == BodyType.STATIC) {
				linearVelocity = Vector2d.ZERO
				angularVelocity = 0.0
				sweep.a0 = sweep.a
				sweep.c0 = sweep.c
				flags = BodyFlags.AWAKE.not(flags)
				synchronizeFixtures()
			}

			isAwake = true

			force = Vector2d.ZERO
			torque = 0.0

			var edge = contactEdge

			while (edge != null) {
				// TODO: проверить, что делает destroy
				world.contactManager.destroy(edge.contact)
				edge = edge.next
			}

			contactEdge = null

			val broadPhase = world.contactManager.broadPhase
			var f: IFixture? = fixtureList

			while (f != null) {
				for (proxy in f.proxies) {
					broadPhase.touchProxy(proxy.proxyId)
				}

				f = f.next
			}
		}

	override fun createFixture(def: FixtureDef): IFixture {
		if (world.isLocked)
			throw ConcurrentModificationException()

		val fixture = Fixture(this, def)

		if (isEnabled) {
			fixture.createProxies(world.contactManager.broadPhase, transform)
		}

		fixture.next = fixtureList
		fixtureList = fixture
		fixtureCount++

		// Adjust mass properties if needed.
		if (fixture.density > 0.0) {
			resetMassData()
		}

		// Let the world know we have a new fixture. This will cause new contacts
		// to be created at the beginning of the next time step.
		world.notifyNewContacts()

		return fixture
	}

	override fun destroyFixture(fixture: IFixture) {
		if (world.isLocked)
			throw ConcurrentModificationException()

		require(fixture.body == this) { "$fixture does not belong to $this (belongs to ${fixture.body})" }
		check(fixtureCount > 0) { "Having no tracked fixtures, but $fixture belongs to us" }

		var node: IFixture? = fixtureList
		var found = false
		var previous: IFixture? = null

		while (node != null) {
			if (node == fixture) {
				// TODO: Это должно работать
				(previous as Fixture?)?.next = node.next
				found = true
				break
			} else {
				previous = node
				node = node.next
			}
		}

		check(found) { "Can't find $fixture in linked list of fixtures" }

		val density = node!!.density

		// Destroy any contacts associated with the fixture.
		var edge = contactEdge

		while (edge != null) {
			val contact = edge.contact
			edge = edge.next

			val fixtureA = contact.fixtureA
			val fixtureB = contact.fixtureB

			if (fixtureA == fixture || fixtureB == fixture) {
				// This destroys the contact and removes it from
				// this body's contact list.
				world.contactManager.destroy(contact)
			}
		}

		fixture as Fixture

		if (isEnabled) {
			fixture.destroyProxies(world.contactManager.broadPhase)
		}

		fixture.unlink()
		fixtureCount--

		// Reset the mass data
		if (density > 0.0) {
			resetMassData()
		}
	}

	override fun resetMassData() {
		// Compute mass data from shapes. Each shape has its own density.
		mass = 0.0
		invMass = 0.0
		rotInertia = 0.0
		rotInertiaInv = 0.0
		sweep.localCenter = Vector2d.ZERO

		if (type == BodyType.STATIC || type == BodyType.KINEMATIC) {
			sweep.c0 = transform.position
			sweep.c = transform.position
			sweep.a0 = sweep.a
			return
		}

		check(type == BodyType.DYNAMIC)

		// Accumulate mass over all fixtures.
		var localCenter = Vector2d.ZERO

		for (fixture in fixtureIterator) {
			if (fixture.density == 0.0) {
				continue
			}

			val massData = fixture.getMassData()
			mass += massData.mass
			localCenter += massData.center * massData.mass
			rotInertia += massData.inertia
		}

		// Compute center of mass.
		if (mass > 0.0) {
			invMass = 1.0 / mass
			localCenter *= invMass
		}

		if (rotInertia > 0.0 && !isFixedRotation) {
			// Center the inertia about the center of mass.
			rotInertia -= mass * localCenter.dotProduct(localCenter)
			check(rotInertia > 0.0)
			rotInertiaInv = 1.0 / rotInertia
		} else {
			rotInertia = 0.0
			rotInertiaInv = 0.0
		}

		// Move center of mass.
		val oldCenter = sweep.c
		sweep.localCenter = localCenter
		sweep.c = transform.times(localCenter)
		sweep.c0 = sweep.c

		// Update center of mass velocity.
		linearVelocity += b2Cross(angularVelocity, sweep.c - oldCenter)
	}

	override var massData: MassData
		get() = MassData(
			mass = mass,
			inertia = inertia,
			center = sweep.localCenter
		)
		set(value) {
			if (world.isLocked)
				throw ConcurrentModificationException()

			if (type != BodyType.DYNAMIC)
				return

			invMass = 0.0
			rotInertia = 0.0
			rotInertiaInv = 0.0

			mass = value.mass

			if (mass <= 0.0) {
				mass = 1.0
			}

			invMass = 1.0 / mass

			if (value.inertia > 0.0 && !isFixedRotation) {
				rotInertia = value.inertia - mass * value.center.dotProduct(value.center)
				check(rotInertia > 0.0)
				rotInertiaInv = 1.0 / rotInertia
			}

			// Move center of mass.
			val oldCenter = sweep.c
			sweep.localCenter = value.center
			sweep.c = transform.times(sweep.localCenter)
			sweep.c0 = sweep.c

			// Update center of mass velocity.
			linearVelocity += b2Cross(angularVelocity, sweep.c - oldCenter)
		}

	internal fun shouldCollide(other: IBody): Boolean {
		// At least one body should be dynamic.
		if (type != BodyType.DYNAMIC && other.type != BodyType.DYNAMIC)
			return false

		// Does a joint prevent collision?
		var joint = jointList

		while (joint != null) {
			if (joint.otherNullable == other && !joint.joint.collideConnected) {
				return false
			}

			joint = joint.next
		}

		return true
	}

	init {
		sweep.c0 = transform.position
		sweep.c = transform.position
		sweep.a0 = def.angle
		sweep.a = def.angle
	}

	override fun setTransform(position: Vector2d, angle: Double) {
		if (world.isLocked)
			throw ConcurrentModificationException()

		transform.rotation.set(angle)
		transform.position = position

		sweep.c = transform.times(sweep.localCenter)
		sweep.a = angle

		sweep.c0 = sweep.c
		sweep.a0 = angle

		val broadPhase = world.contactManager.broadPhase

		for (fixture in fixtureIterator) {
			(fixture as Fixture?)?.synchronize(broadPhase, transform, transform)
		}

		// Check for new contacts the next step
		world.notifyNewContacts()
	}

	override val inertia: Double
		get() = rotInertia + mass * sweep.localCenter.dotProduct(sweep.localCenter)

	override val localCenter: Vector2d
		get() = sweep.localCenter

	override val worldCenter: Vector2d
		get() = sweep.c

	override fun getWorldPoint(localPoint: Vector2d): Vector2d {
		return transform.times(localPoint)
	}

	override fun getWorldVector(localPoint: Vector2d): Vector2d {
		return transform.rotation.times(localPoint)
	}

	override fun getLocalPoint(worldPoint: Vector2d): Vector2d {
		return transform.timesT(worldPoint)
	}

	override fun getLocalVector(worldVector: Vector2d): Vector2d {
		return transform.rotation.timesT(worldVector)
	}

	override fun getLinearVelocityFromWorldPoint(worldPoint: Vector2d): Vector2d {
		return linearVelocity + b2Cross(angularVelocity, worldPoint - sweep.c)
	}

	override fun getLinearVelocityFromLocalPoint(localPoint: Vector2d): Vector2d {
		return getLinearVelocityFromWorldPoint(getWorldPoint(localPoint))
	}

	override fun applyForce(force: Vector2d, point: Vector2d, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		// Don't accumulate a force if the body is sleeping.
		if (isAwake) {
			this.force += force
			this.torque += b2Cross(point - sweep.c, force)
		}
	}

	override fun applyForceToCenter(force: Vector2d, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		// Don't accumulate a force if the body is sleeping.
		if (isAwake) {
			this.force += force
		}
	}

	override fun applyTorque(torque: Double, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		if (isAwake) {
			this.torque += torque
		}
	}

	override fun applyLinearImpulse(impulse: Vector2d, point: Vector2d, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		if (isAwake) {
			linearVelocity += impulse * invMass
			angularVelocity += rotInertiaInv *  b2Cross(point - sweep.c, impulse)
		}
	}

	override fun applyLinearImpulseToCenter(impulse: Vector2d, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		if (isAwake) {
			linearVelocity += impulse * invMass
		}
	}

	override fun applyAngularImpulse(impulse: Double, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		if (isAwake) {
			angularVelocity += impulse * rotInertiaInv
		}
	}

	internal fun synchronizeTransform() {
		transform.rotation.set(sweep.a)
		transform.position = sweep.c - transform.rotation.times(sweep.localCenter)
	}

	internal fun advance(alpha: Double) {
		sweep.advance(alpha)
		sweep.c = sweep.c0
		sweep.a = sweep.a0
		synchronizeTransform()
	}

	internal fun synchronizeFixtures() {
		val broadPhase = world.contactManager.broadPhase

		if (isAwake) {
			val transform1 = Transform()
			transform1.rotation.set(sweep.a0)
			transform1.position = sweep.c0 - transform1.rotation.times(sweep.localCenter)

			for (fixture in fixtureIterator) {
				(fixture as Fixture).synchronize(broadPhase, transform1, transform)
			}
		} else {
			for (fixture in fixtureIterator) {
				(fixture as Fixture).synchronize(broadPhase, transform, transform)
			}
		}
	}

	override fun dump() {
		TODO("Not yet implemented")
	}
}
