package ru.dbotthepony.kbox2d.dynamics.contact

import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.collision.WorldManifold
import ru.dbotthepony.kbox2d.collision.b2TestOverlap
import ru.dbotthepony.kbox2d.dynamics.Body

fun interface ContactFactory {
	fun factorize(fixtureA: IFixture, childIndexA: Int, fixtureB: IFixture, childIndexB: Int): AbstractContact
}

sealed class AbstractContact(
	final override val fixtureA: IFixture,
	final override val childIndexA: Int,
	final override val fixtureB: IFixture,
	final override val childIndexB: Int,
) : IContact {
	internal var flags: Int = ContactFlags.ENABLED.bitmask
	internal var isOnIsland = false
	internal var toiFlag = false

	final override var manifold: Manifold = Manifold()
		private set

	override var next: IContact? = null
		internal set

	override var prev: IContact? = null
		internal set

	internal var isFlaggedForFiltering: Boolean
		get() = ContactFlags.FILTER.isit(flags)
		set(value) { flags = ContactFlags.FILTER.update(flags, value) }

	internal val nodeA: ContactEdge = ContactEdge(contact = this, other = fixtureB.body!!)
	internal val nodeB: ContactEdge = ContactEdge(contact = this, other = fixtureA.body!!)

	init {
		val bodyA = fixtureA.body as Body
		val bodyB = fixtureB.body as Body

		nodeA.next = bodyA.contactEdge
		nodeB.next = bodyB.contactEdge

		bodyA.contactEdge?.prev = nodeA
		bodyB.contactEdge?.prev = nodeB

		bodyA.contactEdge = nodeA
		bodyB.contactEdge = nodeB
	}

	override var friction: Double = b2MixFriction(fixtureA.friction, fixtureB.friction)
	override var restitution: Double = b2MixRestitution(fixtureA.restitution, fixtureB.restitution)
	override var restitutionThreshold: Double = b2MixRestitutionThreshold(fixtureA.restitutionThreshold, fixtureB.restitutionThreshold)
	override var tangentSpeed: Double = 0.0

	internal var toiCount: Int = 0
	internal var toi: Double = 0.0

	override val worldManifold: IWorldManifold get() {
		val bodyA = checkNotNull(fixtureA.body) { "FixtureA has no body attached" }
		val bodyB = checkNotNull(fixtureB.body) { "FixtureB has no body attached" }

		val shapeA = fixtureA.shape
		val shapeB = fixtureB.shape

		return WorldManifold(manifold, bodyA.transform, shapeA.radius, bodyB.transform, shapeB.radius)
	}

	override var isTouching: Boolean
		get() = (flags and ContactFlags.TOUCHING.bitmask) == ContactFlags.TOUCHING.bitmask
		internal set(value) { flags = ContactFlags.TOUCHING.update(flags, value) }

	override var isEnabled: Boolean
		get() = (flags and ContactFlags.ENABLED.bitmask) == ContactFlags.ENABLED.bitmask
		set(value) { flags = ContactFlags.ENABLED.update(flags, value) }

	override fun flagForFiltering() {
		flags = flags or ContactFlags.FILTER.bitmask
	}

	/**
	 * Awakens fixtures this contact belongs to if neither of them are sensors
	 */
	fun destroy() {
		if (manifold.points.isNotEmpty() && !fixtureA.isSensor && !fixtureB.isSensor) {
			fixtureA.body!!.isAwake = true
			fixtureB.body!!.isAwake = true
		}
	}

	/**
	 * Update the contact manifold and touching status.
	 * Note: do not assume the fixture AABBs are overlapping or are valid.
	 */
	internal fun update(contactListener: IContactListener?) {
		// Re-enable this contact.
		flags = ContactFlags.ENABLED.or(flags)

		val oldManifold = manifold

		val touching: Boolean
		val wasTouching = isTouching

		val sensor = fixtureA.isSensor || fixtureB.isSensor

		val bodyA = fixtureA.body as Body
		val bodyB = fixtureB.body as Body

		val xfA = bodyA.transform
		val xfB = bodyB.transform

		// Is this contact a sensor?
		if (sensor) {
			val shapeA = fixtureA.shape
			val shapeB = fixtureB.shape

			touching = b2TestOverlap(shapeA, childIndexA, shapeB, childIndexB, xfA, xfB)
			// Sensors don't generate manifolds.
			manifold = Manifold.EMPTY
		} else {
			manifold = evaluate(xfA, xfB)
			touching = manifold.points.isNotEmpty()

			// Match old contact ids to new contact ids and copy the
			// stored impulses to warm start the solver.
			for (mp2 in manifold.points) {
				mp2.normalImpulse = 0.0
				mp2.tangentImpulse = 0.0
				val id2 = mp2.id

				for (mp1 in oldManifold.points) {
					if (mp1.id.key == id2.key) {
						mp2.normalImpulse = mp1.normalImpulse
						mp2.tangentImpulse = mp1.tangentImpulse
						break
					}
				}
			}

			if (touching != wasTouching) {
				bodyA.isAwake = true
				bodyB.isAwake = true
			}
		}

		this.isTouching = touching

		if (!wasTouching && touching) {
			contactListener?.beginContact(this)
		} else if (wasTouching && !touching) {
			contactListener?.endContact(this)
		}

		if (!sensor && touching) {
			contactListener?.preSolve(this, oldManifold)
		}
	}

	companion object {
		private val registry =
			Object2ObjectArrayMap<IShape.Type, Object2ObjectArrayMap<IShape.Type, ContactFactory>>()

		internal fun register(type1: IShape.Type, type2: IShape.Type, factory: ContactFactory) {
			registry.computeIfAbsent(type1, Object2ObjectFunction { Object2ObjectArrayMap() })
				.put(type2, factory)
		}

		internal fun create(
			fixtureA: IFixture,
			indexA: Int,
			fixtureB: IFixture,
			indexB: Int,
		): AbstractContact {
			val type1 = fixtureA.type
			val type2 = fixtureB.type

			return registry[type1]?.get(type2)?.factorize(fixtureA, indexA, fixtureB, indexB)
				?: registry[type2]?.get(type1)?.factorize(fixtureB, indexB, fixtureA, indexA)
				?: throw IllegalArgumentException("No collision handler for between $type1 and $type2")
		}

		init {
			register(IShape.Type.POLYGON, IShape.Type.POLYGON) { fixtureA: IFixture, _: Int, fixtureB: IFixture, _: Int ->
				return@register PolygonContact(fixtureA, fixtureB)
			}

			register(IShape.Type.POLYGON, IShape.Type.CIRCLE) { fixtureA: IFixture, _: Int, fixtureB: IFixture, _: Int ->
				return@register PolygonCircleContact(fixtureA, fixtureB)
			}

			register(IShape.Type.CIRCLE, IShape.Type.CIRCLE) { fixtureA: IFixture, _: Int, fixtureB: IFixture, _: Int ->
				return@register CircleContact(fixtureA, fixtureB)
			}

			register(IShape.Type.EDGE, IShape.Type.CIRCLE) { fixtureA: IFixture, _: Int, fixtureB: IFixture, _: Int ->
				return@register EdgeCircleContact(fixtureA, fixtureB)
			}

			register(IShape.Type.EDGE, IShape.Type.POLYGON) { fixtureA: IFixture, _: Int, fixtureB: IFixture, _: Int ->
				return@register EdgePolygonContact(fixtureA, fixtureB)
			}

			register(IShape.Type.CHAIN, IShape.Type.POLYGON) { fixtureA: IFixture, indexA: Int, fixtureB: IFixture, indexB: Int ->
				return@register ChainPolygonContact(fixtureA, indexA, fixtureB, indexB)
			}

			register(IShape.Type.CHAIN, IShape.Type.CIRCLE) { fixtureA: IFixture, indexA: Int, fixtureB: IFixture, indexB: Int ->
				return@register ChainCircleContact(fixtureA, indexA, fixtureB, indexB)
			}
		}
	}
}
