package ru.dbotthepony.kbox2d.dynamics.joint

import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.dynamics.Body
import ru.dbotthepony.kstarbound.math.Vector2d
import kotlin.math.PI

fun interface JointFactory {
	fun factorize(jointDef: IJointDef): AbstractJoint
}

sealed class AbstractJoint(def: IJointDef) : IJoint {
	init {
		require(def.bodyA != def.bodyB) { "Tried to create join on same body" }
	}

	internal var isOnIsland = false

	protected var index: Int = 0

	override val collideConnected: Boolean = def.collideConnected
	override val type: JointType = def.type
	override var userData: Any? = def.userData

	// KBox2D: In original code, nothing expects bodies to be null
	// but certain joints (notably, mouse joint) have no meaningful
	// value for second body involved. So, KBox2D CAN handle case where
	// joint have only one body, and to avoid null assertions in all places
	// possible, bodyA and bodyB getters assert for null by themselves
	// However, there is nullable getter for bodies: nullableBodyA and nullableBodyB
	protected var _bodyA: Body? = def.bodyA as Body?
	protected var _bodyB: Body? = def.bodyB as Body?

	final override val bodyA: Body get() = checkNotNull(_bodyA) { "Body A is not present" }
	final override val bodyB: Body get() = checkNotNull(_bodyB) { "Body B is not present" }

	val hasBodyA: Boolean get() = _bodyA != null
	val hasBodyB: Boolean get() = _bodyB != null
	val hasTwoBodies: Boolean get() = hasBodyA && hasBodyB

	val nullableBodyA get() = _bodyA
	val nullableBodyB get() = _bodyB

	private var _edgeA: JointEdge? = JointEdge(other = _bodyB, joint = this, next = _bodyA?.jointList)
	private var _edgeB: JointEdge? = JointEdge(other = _bodyA, joint = this, next = _bodyB?.jointList)

	internal val edgeA: JointEdge get() = checkNotNull(_edgeA) { "Edge A is not present" }
	internal val edgeB: JointEdge get() = checkNotNull(_edgeB) { "Edge B is not present" }

	var isValid: Boolean = true
		private set

	init {
		// Connect to the bodies' doubly linked lists.
		_bodyA?.jointList?.prev = edgeA
		_bodyA?.jointList = edgeA
		_bodyB?.jointList?.prev = edgeB
		_bodyB?.jointList = edgeB
	}

	final override var next: IJoint? = null
		internal set
	final override var prev: IJoint? = null
		internal set

	/**
	 * Signals that this joint was destroyed, invalidate stuff
	 * to fail-fast this object
	 */
	internal open fun unlink() {
		_edgeA = null
		_edgeB = null
		_bodyA = null
		_bodyB = null
		isValid = false
	}

	internal abstract fun initVelocityConstraints(data: B2SolverData)
	internal abstract fun solveVelocityConstraints(data: B2SolverData)
	internal abstract fun solvePositionConstraints(data: B2SolverData): Boolean

	override fun shiftOrigin(newOrigin: Vector2d) {
		// no-op
	}

	companion object {
		private val registry = Object2ObjectArrayMap<JointType, JointFactory>()

		internal fun register(jointType: JointType, factory: JointFactory) {
			require(registry.put(jointType, factory) == null) { "Re-registered $jointType factory" }
		}

		internal fun create(jointType: JointType, jointDef: IJointDef): AbstractJoint {
			return requireNotNull(registry[jointType]) { "No joint factory registered for type $jointType" }.factorize(jointDef)
		}

		internal fun create(jointDef: IJointDef): AbstractJoint {
			return create(jointDef.type, jointDef)
		}

		init {
			register(JointType.DISTANCE) {
				return@register DistanceJoint(it as DistanceJointDef)
			}

			register(JointType.REVOLUTE) {
				return@register RevoluteJoint(it as RevoluteJointDef)
			}

			register(JointType.PRISMATIC) {
				return@register PrismaticJoint(it as PrismaticJointDef)
			}

			register(JointType.PULLEY) {
				return@register PulleyJoint(it as PulleyJointDef)
			}

			register(JointType.GEAR) {
				return@register GearJoint(it as GearJointDef)
			}

			register(JointType.MOUSE) {
				return@register MouseJoint(it as MouseJointDef)
			}

			register(JointType.WHEEL) {
				return@register WheelJoint(it as WheelJointDef)
			}

			register(JointType.WELD) {
				return@register WeldJoint(it as WeldJointDef)
			}

			register(JointType.FRICTION) {
				return@register FrictionJoint(it as FrictionJointDef)
			}

			register(JointType.MOTOR) {
				return@register MotorJoint(it as MotorJointDef)
			}
		}
	}
}
